(**********************************************************************)
(*                                                                    *)
(*                           Moca                                     *)
(*                                                                    *)
(*          Pierre Weis, INRIA Rocquencourt                           *)
(*          Frdric Blanqui, projet Protheo, INRIA Lorraine          *)
(*                                                                    *)
(*  Copyright 2005-2008,                                              *)
(*  Institut National de Recherche en Informatique et en Automatique. *)
(*  All rights reserved.                                              *)
(*                                                                    *)
(*  This file is distributed under the terms of the Q Public License. *)
(*                                                                    *)
(**********************************************************************)

(* result type for comparison functions *)

type ord =
  | Greater
  | Equal
  | Smaller
  | Uncomparable;;

type 'a cmp = 'a -> 'a -> ord;;

(* basic functions on ord *)

let opp = function
  | Greater -> Smaller
  | Smaller -> Greater
  | (Equal | Uncomparable) as x -> x;;

let ge = function
  | Greater | Equal -> true
  | Smaller | Uncomparable -> false;;

let le = function
  | Smaller | Equal -> true
  | Greater | Uncomparable -> false;;

let comp o1 o2 =
  match o1, o2 with
    | Smaller, Smaller | Smaller, Equal | Equal, Smaller
      -> Smaller
    | Equal, Equal
      -> Equal
    | Greater, Greater | Greater, Equal | Equal, Greater
      -> Greater
    | Greater, Smaller | Smaller, Greater | Uncomparable, Uncomparable
    | Uncomparable, Smaller | Uncomparable, Equal | Uncomparable, Greater
    | Smaller, Uncomparable | Equal, Uncomparable | Greater, Uncomparable
      -> Uncomparable;;

let int_of_ord = function
  | Greater -> 1
  | Equal -> 0
  | Smaller -> -1
  | Uncomparable -> raise (Invalid_argument "inf_of_ord");;

let gt_if = function
  | true -> Greater
  | false -> Uncomparable;;

let lt_if = function
  | true -> Smaller
  | false -> Uncomparable;;

let eq_if = function
  | true -> Equal
  | false -> Uncomparable;;

(* lexicographic extension of an ordering *)

let lex cmp =
  let rec aux xs ys =
    match xs, ys with
      | [], [] -> Equal
      | x :: xs2, y :: ys2 ->
	  (match cmp x y with
	     | Equal -> aux xs2 ys2
	     | (Smaller | Greater | Uncomparable) as n -> n)
      | _ -> Uncomparable
  in aux;;

(* multiset extension of an ordering *)
(* See Baader and Nipkow's Term Rewriting and All That, p.27 *)

let mul cmp =
  let rec rem xs y =
     match xs with
     | [] -> xs
     | x :: xs1 ->
       match cmp x y with
       | Equal -> xs1
       | Greater | Smaller | Uncomparable -> x :: rem xs1 y in 
   let rec mdiff xs ys =
      match ys with 
      | [] -> xs
      | y :: ys1 ->
        mdiff (rem xs y) ys1 in
   let mul_is_greater e1 e2 =
    List.for_all 
       (fun n -> 
        List.exists (fun m -> (cmp m n) = Greater) e1) e2 in
   let aux m n =
    let n_m = mdiff n m 
    and m_n = mdiff m n in
    match n_m, m_n with
    | [], [] -> Equal
    | _, _ ->
     if mul_is_greater m_n n_m
     then Greater
     else 
     if mul_is_greater n_m m_n then Smaller
     else Uncomparable 
   in aux     
;;

(* lexicographic path ordering *)

open Symb;;
open Term;;
open Subterm;;
open Parsetree;;

exception IncompletePrecedence of symb * symb;;

let lpo_fail cmp =
  let rec aux t u =
    match t, u with
      | (App (f,ts) as t), (App (g,us) as u) ->
	  if List.exists (fun ti -> ge (aux ti u)) ts then Greater
	  else if List.exists (fun ui -> le (aux t ui)) us then Smaller
	  else
	    (match cmp f g with
	       | Equal -> lex aux ts us
	       | Greater -> gt_if (List.for_all
                                     (fun ui -> aux t ui = Greater) us)
	       | Smaller -> lt_if (List.for_all
                                     (fun ti -> aux ti u = Smaller) ts)
	       | Uncomparable -> raise (IncompletePrecedence (f,g)))
      | Var x, Var y -> eq_if (x=y)
      | App _ as t, Var x -> gt_if (occurs x t)
      | Var x, (App _ as t) -> lt_if (occurs x t)
  in aux;;

let lpo cmp t u =
  try lpo_fail cmp t u
  with IncompletePrecedence _ -> Uncomparable;;

let mpo_fail cmp =
  let rec aux t u =
    match t, u with
      | (App (f, ts) as t), (App (g, us) as u) ->
	  if List.exists (fun ti -> ge (aux ti u)) ts then Greater
	  else if List.exists (fun ui -> le (aux t ui)) us then Smaller
	  else
	    (match cmp f g with
	       | Equal -> mul aux ts us
	       | Greater -> gt_if (List.for_all
                                     (fun ui -> aux t ui = Greater) us)
	       | Smaller -> lt_if (List.for_all
                                     (fun ti -> aux ti u = Smaller) ts)
	       | Uncomparable -> raise (IncompletePrecedence (f,g)))
      | Var x, Var y -> eq_if (x=y)
      | App _ as t, Var x -> gt_if (occurs x t)
      | Var x, (App _ as t) -> lt_if (occurs x t)
  in aux;;

let mpo cmp t u =
  try mpo_fail cmp t u
  with IncompletePrecedence _ -> Uncomparable;;

let rec rpo_fail status cmp =
let rec get_status f = 
 try Hashtbl.find status f
 with Not_found -> Lex
in
  let rec aux t u =
    match t, u with
    | Var x , Var y -> eq_if (x=y)
    | (App _ as t), Var y -> gt_if (occurs y t) 
    | Var x, (App _ as u) -> lt_if (occurs x u)
    | (App (f, ts) as t), (App (g, us) as u) ->
       if List.exists (fun ti -> ge (aux ti u)) ts
       then Greater
       else if List.exists (fun ui -> le (aux t ui)) us
       then Smaller
       else ( 
        match cmp f g with
       | Greater -> gt_if (List.for_all
                             (fun ui -> aux t ui = Greater ) us)
       | Smaller -> lt_if (List.for_all
                             (fun ti -> aux ti u = Smaller) ts)
       | Equal ->
       ( match get_status f, get_status g with
        | Mul, Mul -> mul aux ts us
        | Lex, Lex -> lex aux ts us
        | Mul, Lex | Lex, Mul -> assert false
        )      
       | Uncomparable -> raise (IncompletePrecedence (f, g)))
in aux;;


let rpo status cmp t u =
  try rpo_fail status cmp t u
  with IncompletePrecedence _ -> Uncomparable
;;

