(**********************************************************************)
(*                                                                    *)
(*                           Moca                                     *)
(*                                                                    *)
(*          Laura Lowenthal, projet Protheo, INRIA Lorraine           *)
(*                                                                    *)
(*  Copyright 2005-2007,                                              *)
(*  Institut National de Recherche en Informatique et en Automatique. *)
(*  All rights reserved.                                              *)
(*                                                                    *)
(*  This file is distributed under the terms of the Q Public License. *)
(*                                                                    *)
(**********************************************************************)

(* $Id: genr_varyadic_equations.ml,v 1.7 2008/02/13 10:58:59 lowenthl Exp $ *)

(** {Generation of test equations for varyadic constructors.} *)

open Parsetree
open Term
open Term_utils
open Funutils
;;

let mk_rel desc = {prel_loc = Location.none; prel_desc = desc};;

let rec testing_equations_of_rel g rel =
  match rel.prel_desc with
    | Rewrite _  -> Eqnrel.eqns_of_rel g rel
    | Commutative _ -> [
        (app g [x; y; z; w], app g [y; x; w; z]);
        (app g [x; y; z; w], app g [w; y; z; x]);
        (app g [x; y; z; w], app g [w; z; y; x]);
        (app g [x; y; z; w], app g [z; y; x; w])
      ]
    | Associative _ -> [
        (app g [app g [x; y]; z], app g [x; y; z]);
        (app g [x; app g [y; z] ; app g [w; Var 4]], app g [x; y; z; w; Var 4]);
        (app g [app g [x; y; z]; w], app g [x; y; z; w])
      ]
    | Absorbing _ -> []
    | Neutral (Left, e) -> let e = app0 e in [
        (app g [x; e; y], app g [x; y]);
        (app g [e; x], x);
        (app g [x; e; e], app g [x; e]);
        (app g [x; e; e; y], app g [x; y]);
      ]
    | Neutral (Right, e) -> let e = app0 e in [
        (app g [x; e; y], app g [x; y]);
        (app g [x; e], x);
        (app g [e; e; y], app g [e; y]);
        (app g [x; e; e; y], app g [x; y]);
      ]
    | Neutral (Both, e) -> let e = app0 e in [
        (app g [x; e; y], app g [x; y]);
        (app g [e; x], x);
        (app g [x; e; e], x);
      ]
    | Idempotent Left -> [
        (app g [x; x; y], app g [x; y]);
        (app g [x; y; y; z], app g [x; y; z])
      ]
    | Idempotent Right -> [
        (app g [y; x; x], app g [y; x]);
        (app g [x; y; y; z], app g [x; y; z])
      ]
    | Idempotent Both -> [
        (app g [x; x], x);
        (app g [x; y; y; z], app g [x; y; z])
      ]
    | Nilpotent (Left, e) -> let e = app0 e in [
        (app g [x; x; y], app g [e; y]);
        (app g [x; y; y; z], app g [x; e; z])
      ]
    | Nilpotent (Right, e) -> let e = app0 e in [
        (app g [y; x; x], app g [y; e]);
        (app g [x; y; y; z], app g [x; e; z])
      ]
    | Nilpotent (Both, e) -> let e = app0 e in [
        (app g [x; x], e);
        (app g [x; y; y; z], app g [x; e; z])
      ]
    | Involutive -> []
    | Inverse (_, _, None) -> []
    | Inverse (Left, i, Some e) -> let e = app0 e in [
        (app g [app1 i x; x], e);
        (app g [x; app1 i y; y; z], app g [x; e; z]) 
      ]
    | Inverse (Right, i, Some e) -> let e = app0 e in [
        (app g [x; app1 i x], e);
        (app g [x; y; app1 i y; z], app g [x; e; z]) 
      ]
    | Inverse (Both, i, e) ->
        testing_equations_of_rel g (mk_rel ^$ Inverse (Left, i, e)) @
        testing_equations_of_rel g (mk_rel ^$ Inverse (Right, i, e))
    | Distributive (_, d, eopt, dist) ->
        let e = match eopt with Some e -> e | None -> d in
        let n = if Check.is_varyadic_gen d then 3 else Check.arity_gen d in
        let vs = vars n in 
        let args1 = List.map (fun v -> app g [x; y; v; z]) vs in
        let args2 = List.map (fun v -> app g [v; x]) vs in
        let args1, args2 = match dist with
          Dist_Inverse -> List.rev args1, List.rev args2
        | Dist_Direct -> args1, args2 in
        [ 
          (app g [x; y; app d vs; z], app e args1);
          (app g [app d vs; x], app e args2)
        ]
    | Division_by_Absorbent _
    | Structure_item _
    | Status _
    | Precedence _ -> []

let testing_equations_of_rels c rels =
  match rels.prels_desc with
    | Prels_none
    | Prels_commented _ -> []
    | Prels_begend rs -> Listutils.flat_map (testing_equations_of_rel c) rs
