(**********************************************************************)
(*                                                                    *)
(*                           Moca                                     *)
(*                                                                    *)
(*          Laura Lowenthal, projet Protheo, INRIA Lorraine           *)
(*                                                                    *)
(*  Copyright 2005-2007,                                              *)
(*  Institut National de Recherche en Informatique et en Automatique. *)
(*  All rights reserved.                                              *)
(*                                                                    *)
(*  This file is distributed under the terms of the Q Public License. *)
(*                                                                    *)
(**********************************************************************)

(* $Id: tgenr_base.ml,v 1.10 2008/02/13 10:58:59 lowenthl Exp $ *)

(** {Basic functions for value generation} *)

open Code
open Parsetree
open Funutils
open Listutils

(** Creates an expression in Code.exp representing a constant *)
let mk_constant constant = Ocaml_expression ({
   pexp_desc = Pexp_constant (constant);
   pexp_loc = Location.none;
 })
let mk_int n = mk_constant (Asttypes.Const_int n)
let mk_char ch = mk_constant (Asttypes.Const_char ch)
let mk_string str = mk_constant (Asttypes.Const_string str)
let mk_float x = mk_constant (Asttypes.Const_float (string_of_float x))
let mk_int32 n = mk_constant (Asttypes.Const_int32 n)
let mk_int64 n = mk_constant (Asttypes.Const_int64 n)
let mk_nativeint n = mk_constant (Asttypes.Const_nativeint n)
;;

(** Creates an expression in Code.exp representing a tuple *)
let mk_tuple exps = Tuple exps
;;

(** Creates an expression in Code.exp representing a list *)
let mk_list exps = List exps
;;

(** Creates a generator with the given name and location *)
let mk_generator name loc =
  {pgen_desc = Longident.Lident name; pgen_loc = loc}
;;

let mk_non_varyadic_construction name args =
  let gen_function_name = Genr_base.construction_function_name name in
  let gen_function_ident = make_Constant gen_function_name in
  match args with
    [] -> gen_function_ident
  | [_] -> Apply (gen_function_ident, args)
  | _ -> Apply (gen_function_ident, [Tuple args])
;;

let mk_varyadic_construction name args =
  let gen_function_name = Genr_base.construction_function_name name in
  let gen_function_ident = make_Constant gen_function_name in
  Apply (gen_function_ident, [List args])
;;

let is_known_and_varyadic gen args =
  (try Check.is_varyadic_gen gen with
    | Check.Error _  -> false) &&
  List.length args > 1
;;

let is_known_and_varyadic_name gname args =
  is_known_and_varyadic (mk_generator gname Location.none) args
;;

(** Creates an expression in Code.exp representing a generator with the *)
 (** specified name and arguments *)
let mk_construction name args =
  let gen = mk_generator name Location.none in
    if is_known_and_varyadic gen args
    then mk_varyadic_construction name args
    else mk_non_varyadic_construction name args
;;

(** Creates an expression representing the equality between two arguments *)
let mk_equality e1 e2 = infix_name e1 "=" e2
;;

let choose_ints rnd n =
  if rnd then Randomutils.ints n (n * 5)
  else Listutils.from_to 1 n succ

(** Creates a number of Code.expressions respresenting distinct constants *)
let genr_constants_from_ints mk rnd n = List.map mk (choose_ints rnd n)
let genr_ints = genr_constants_from_ints mk_int
let genr_chars = genr_constants_from_ints (mk_char <.> char_of_int)
let genr_strings = genr_constants_from_ints (mk_string <.> string_of_int)
let genr_floats = genr_constants_from_ints (mk_float <.> float_of_int)
let genr_int32s = genr_constants_from_ints (mk_int32 <.> Int32.of_int)
let genr_int64s = genr_constants_from_ints (mk_int64 <.> Int64.of_int)
let genr_nativeints = genr_constants_from_ints (mk_nativeint <.>
 Nativeint.of_int)
let genr_units n = Listutils.take n [mk_tuple []]
;;

