(**********************************************************************)
(*                                                                    *)
(*                           Moca                                     *)
(*                                                                    *)
(*          Laura Lowenthal, projet Protheo, INRIA Lorraine           *)
(*                                                                    *)
(*  Copyright 2005-2007,                                              *)
(*  Institut National de Recherche en Informatique et en Automatique. *)
(*  All rights reserved.                                              *)
(*                                                                    *)
(*  This file is distributed under the terms of the Q Public License. *)
(*                                                                    *)
(**********************************************************************)

(* $Id: typed_vars.ml,v 1.9 2008/02/13 10:58:59 lowenthl Exp $ *)

(** {Deducing mapping typed variables to values Code.exp.} *)

open Var;;
open Term;;
open T_otype;;
open Parsetree;;
open Parsetree_utils;;
open Symb;;

type typed_var = var * core_type
;;

exception Unknown_type
;;

let eq_typed_var (v1, t1) (v2, t2) = v1 = v2 && eq_type t1 t2
;;

let gen_info cname = Check.find_gname (string_of_symbol cname)
;;

let is_varyadic cname args =
  (try Check.is_varyadic_info (gen_info cname) with
    | Not_found -> false) &&
  List.length args > 1
;;

let known_general_type cname =
  let mk_core_type td = {ptyp_loc = Location.none;
      ptyp_desc = td} in
  let list_ident = Longident.Lident "list" in
  let list_type arg = Ptyp_constr (list_ident, [mk_core_type arg]) in
  let tvar_a = Ptyp_var "a" in
  let type_descs = match (string_of_symbol cname) with
    | "[]" -> []
    | "::" -> [tvar_a; list_type tvar_a]
    | _ -> raise Not_found in
  ("list", ["a"], List.map mk_core_type type_descs)
;;

let general_type cname =
  try
   let c_info = gen_info cname in
   let type_name = c_info.Check.gi_type_name in
   let params = (Check.find_tname type_name).ptype_params in
   (type_name, params, c_info.Check.gi_arg_types) with
   | Not_found -> try known_general_type cname with
     | Not_found -> raise Unknown_type
;;

let rec unif_arg_types topt (tname, params, arg_types) =
  match topt with
    | None -> arg_types
    | Some core_type ->
        match core_type.ptyp_desc with
          | Ptyp_any | Ptyp_var _ | Ptyp_arrow _ | Ptyp_tuple _
          | Ptyp_object _ | Ptyp_class _ | Ptyp_alias _
          | Ptyp_variant _ | Ptyp_poly _ -> arg_types
          | Ptyp_constr (real_name, real_args) ->
              let binding = build_binding real_name real_args tname params in
              subst binding arg_types
          | Ptyp_parens t -> unif_arg_types (Some t) (tname, params, arg_types)

and build_binding real_name real_args tname params =
  if real_name = Longident.Lident tname
  then try List.combine params real_args with
    | Invalid_argument _ -> raise Unknown_type
  else raise Unknown_type
;;

let checked_types real_type_opt cname terms =
  let c_type = general_type cname in
  let arg_types = unif_arg_types real_type_opt c_type in
  if is_varyadic cname terms
  then let arg_type = Otype.list_argument (List.hd arg_types) in
    Listutils.repeat (List.length terms) arg_type
  else arg_types

let rec checked_typed_vars topt = function
  | Var _ -> []
  | App (cname, terms) ->
      let terms_types = List.combine terms (checked_types topt cname terms) in
      let aux_typed_vars = function
        | (Var v, t) -> [(v, t)]
        | (App (c, ts), t) -> checked_typed_vars (Some t) (App (c, ts)) in
       Listutils.flat_map aux_typed_vars terms_types

let typed_vars = checked_typed_vars None
;;

let rec typed_vars_eqn (l, r) = Listutils.nub_eq eq_typed_var (typed_vars l @ typed_vars r)
;;

let has_type t1 (_, t2) = eq_type t1 t2
;;

let rec vars_by_type = function
  | [] -> []
  | (v, t) :: vts ->
    let (t_vars, non_t_vars) = List.partition (has_type t) vts in
    (t, v :: List.map fst t_vars) :: vars_by_type non_t_vars
;;
